extends Node2D


const BULLET = preload("res://Scenes/PlayerBullet.tscn")
const GHOST = preload("res://Scenes/Ghost.tscn")
const EXPLOSION = preload("res://Scenes/ShipExplosion.tscn")
const LIMITS = [48, 950, 520, 10]

export var dash_cooldown_time = 3.5
export var dash_duration_time = 0.35
export var dash_speed = 1600
export var ghost_cooldown_time = 0.1
export var ghost_duration_time = 0.5
export var shoot_cooldown_time = 0.1
export var initial_speed = 800
export(NodePath) var GameCamera
export(NodePath) var GameManager
export(NodePath) var DashBar
export(NodePath) var InvertBar
export(NodePath) var LifeBar
export(NodePath) var Speaker

var dash_cooldown = 0
var dash_duration = 0
var ghost_cooldown = 0
var input_vector = Vector2.ZERO
var dash_vector = Vector2.ZERO
var inverted = false
var invert_cooldown = 0
var invert_cooldown_time = 5
var shoot_cooldown = 0
var speed = 0
var state = PlayerState.WAITING
var reset_position = Vector2(480, 590)

var life_points_max = 25
var life_points = 25
var max_damage = 3
var flashing = false
var flash_cooldown = 0.1
var flash_cooldown_time = 0.1

var win_cooldown = 3
var win_cooldown_time = 3
var invencible = false


# Ready ========================================================================
func _ready():
	GameCamera = get_node(GameCamera)
	GameManager = get_node(GameManager)
	Speaker = get_node(Speaker)
	DashBar = get_node(DashBar)
	LifeBar = get_node(LifeBar)
	InvertBar = get_node(InvertBar)
	reset()


# Process ======================================================================
func _process(delta):

	if flashing:
		flash_cooldown -= delta
		if flash_cooldown <= 0:
			flash(false)
	
	if shoot_cooldown > 0:
		shoot_cooldown -= delta
	
	if state == PlayerState.READY:
		
		if (speed - 0.0001) > initial_speed:
			speed = lerp(speed, initial_speed, 0.25)
		
		# Movement
		input_vector.x = Input.get_action_strength("ui_right") - Input.get_action_strength("ui_left")
		input_vector.y = Input.get_action_strength("ui_down") - Input.get_action_strength("ui_up")
		translate(input_vector.normalized() * speed * delta)
		global_position.x = clamp(global_position.x, LIMITS[3], LIMITS[1])
		global_position.y = clamp(global_position.y, LIMITS[0], LIMITS[2])

		# Shoot
		if Input.is_action_pressed("ui_shoot"):
			shoot()
		
		# Dash 
		if Input.is_action_just_pressed("ui_dash"):
			dash()

		# Invert
		if Input.is_action_just_pressed("ui_invert"):
			invert_universe()
	
	if state == PlayerState.DASHING:
		translate(dash_vector.normalized() * dash_speed * delta)
		ghost_cooldown -= delta
		if ghost_cooldown <= 0:
			ghost_cooldown = ghost_cooldown_time
			ghost()
	
	if state != PlayerState.DASHING and state != PlayerState.DEAD:
		if dash_cooldown > 0:
			dash_cooldown -= delta
			DashBar.update_value(dash_cooldown_time - dash_cooldown)

	if state != PlayerState.DEAD:
		if invert_cooldown > 0:
			invert_cooldown -= delta
			InvertBar.update_value(invert_cooldown_time - invert_cooldown)
	
	if state == PlayerState.WIN:
		if global_position.y > -540:
			speed = lerp(speed, dash_speed, 0.1)
			translate(Vector2(0, -1) * speed * delta)
		else:
			explode(true)


# Actions ======================================================================

# Invert -----------------------------------------------------------------------
func invert(is_inverted):
	inverted = is_inverted
	$Sprite.frame = 0 if inverted else 1


# Invert Universe --------------------------------------------------------------
func invert_universe():
	if invert_cooldown <= 0:
		invert_cooldown = invert_cooldown_time
		InvertBar.update_value(0, true, 3)
		GameManager.invert()

# Shoot ------------------------------------------------------------------------
func shoot():
	if shoot_cooldown <= 0:
		var bullet = BULLET.instance()
		get_parent().add_child(bullet)
		bullet.invert(inverted)
		bullet.global_position = $CannonPosition.global_position
		shoot_cooldown = shoot_cooldown_time


# Dash -------------------------------------------------------------------------
func dash():
	if dash_cooldown <= 0 and state != PlayerState.WIN:
		state = PlayerState.DASHING
		ghost_cooldown = 0
		speed = dash_speed
		dash_vector = input_vector if input_vector != Vector2(0, 0) else Vector2(0, -1)
		DashBar.update_value(0, true, 3)
		yield(get_tree().create_timer(dash_duration_time), 'timeout')
		state = PlayerState.READY
		dash_cooldown = dash_cooldown_time


# Ghost ------------------------------------------------------------------------
func ghost():
	var new_ghost = GHOST.instance()
	get_parent().add_child(new_ghost)
	new_ghost.global_position = global_position
	new_ghost.disappear($Sprite.texture, $Sprite.vframes, $Sprite.hframes, $Sprite.frame, ghost_duration_time)


# Flash ------------------------------------------------------------------------
func flash(active):
	flashing = active
	flash_cooldown = flash_cooldown_time
	$Sprite.material.set_shader_param('inverted', inverted)
	$Sprite.material.set_shader_param('active', active)


# Take damage ------------------------------------------------------------------
func take_damage(damage):
	if not invencible:
		life_points -= clamp(damage, 1, max_damage)
		LifeBar.update_value(life_points, true)
		Speaker.play_effect("bullet_impact")
		if life_points > 0:
			flash(true)
			GameCamera.screen_shake(1.5)
		else:
			life_points = 0
			explode()


# Explode ----------------------------------------------------------------------
func explode(winner=false):
	var main_menu = get_parent().get_node("MainMenu")
	if winner:
		state = PlayerState.DEAD
		yield(get_tree().create_timer(win_cooldown), 'timeout')
		main_menu.enter_dead_menu()
	else:
		var explosion = EXPLOSION.instance()
		get_parent().add_child(explosion)
		explosion.global_position = global_position
		explosion.explode(inverted)
		main_menu.enter_dead_menu()
		Speaker.play_voice("you_lose")
		state = PlayerState.DEAD
		visible = false


# Reset ship -------------------------------------------------------------------
func reset():
	state = PlayerState.WAITING
	speed = initial_speed
	dash_cooldown = 0
	dash_duration = 0
	ghost_cooldown = 0
	life_points = life_points_max
	global_position.x = reset_position.x
	global_position.y = reset_position.y
	visible = true
	invencible = false
	LifeBar.reset(life_points_max)
	DashBar.reset(dash_cooldown_time)
	InvertBar.reset(invert_cooldown_time)


# Go ---------------------------------------------------------------------------
func go():
	state = PlayerState.READY


# Win --------------------------------------------------------------------------
func win():
	var win_data_animation = get_parent().get_node("ShipInterface/WinData/Animation")
	invencible = true
	win_cooldown = win_cooldown_time
	yield(get_tree().create_timer(1), 'timeout')
	Speaker.play_voice("you_win")
	yield(get_tree().create_timer(2), 'timeout')
	state = PlayerState.WIN
	GameManager.update_final_score()
	GameManager.inverted = true
	GameManager.invert()
	win_data_animation.play("Enter")


# Signals ======================================================================

# When the player ship has been hited ------------------------------------------
func _on_Area_area_entered(area):
	if state == PlayerState.READY:
		if [2].find(area.collision_layer) >= 0:
			var enemy_ship = area.get_parent()
			take_damage(enemy_ship.life_points)
			enemy_ship.explode()
		if [8].find(area.collision_layer) >= 0:
			var enemy_bullet = area.get_parent()
			if enemy_bullet.inverted == inverted:
				take_damage(1)
				enemy_bullet.explode()
